# -*- coding: utf-8 -*-
#######################################################################
# Name: test_decorator_combine
# Purpose: Test for Combine decorator. Combine decorator
#           results in Terminal parse tree node. Whitespaces are
#           preserved  (they are not skipped) and comments are not matched.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest
from arpeggio import ParserPython, ZeroOrMore, OneOrMore, NonTerminal, \
    Terminal, NoMatch, Combine


def test_combine_python():

    # This will result in NonTerminal node
    def root():
        return my_rule(), "."

    # This will result in Terminal node
    def my_rule():
        return Combine(ZeroOrMore("a"), OneOrMore("b"))

    parser = ParserPython(root)

    input1 = "abbb."

    # Whitespaces are preserved in lexical rules so the following input
    # should not be recognized.
    input2 = "a b bb."

    ptree1 = parser.parse(input1)

    with pytest.raises(NoMatch):
        parser.parse(input2)

    assert isinstance(ptree1, NonTerminal)
    assert isinstance(ptree1[0], Terminal)
    assert ptree1[0].value == "abbb"



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_reduce_tree
# Purpose: Test parse tree reduction
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest  # noqa

# Grammar
from arpeggio import ZeroOrMore, OneOrMore, ParserPython, Terminal, NonTerminal
from arpeggio import RegExMatch as _


def grammar():      return first, "a", second, [first, second]
def first():        return [fourth, third], ZeroOrMore(third)
def second():       return OneOrMore(third), "b"
def third():        return [third_str, fourth]
def third_str():    return "3"
def fourth():       return _(r'\d+')


def test_reduce_tree():

    input = "34 a 3 3 b 3 b"

    parser = ParserPython(grammar, reduce_tree=False)
    result = parser.parse(input)

#    PTDOTExporter().exportFile(result, 'test_reduce_tree_pt.dot')

    assert result[0].rule_name == 'first'
    assert isinstance(result[0], NonTerminal)
    assert result[3].rule_name == 'first'
    assert result[0][0].rule_name == 'fourth'
    # Check reduction for direct OrderedChoice
    assert result[2][0].rule_name == 'third'

    parser = ParserPython(grammar, reduce_tree=True)
    result = parser.parse(input)

    # PTDOTExporter().exportFile(result, 'test_reduce_tree_pt.dot')

    assert result[0].rule_name == 'fourth'
    assert isinstance(result[0], Terminal)
    assert result[3].rule_name == 'fourth'
    # Check reduction for direct OrderedChoice
    assert result[2][0].rule_name == 'third_str'



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_position
# Purpose: Test that positions in the input stream are properly calculated.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2015-2017 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import unicode_literals
import pytest
from arpeggio import ParserPython


@pytest.fixture
def parse_tree():

    def grammar(): return ("first", "second", "third")

    parser = ParserPython(grammar)

    return parser.parse("   first \n\n  second   third")


def test_position(parse_tree):
    assert parse_tree[0].position == 3
    assert parse_tree[1].position == 13
    assert parse_tree[2].position == 22
    assert parse_tree.position == 3


def test_position_end(parse_tree):
    assert parse_tree[0].position_end == 8
    assert parse_tree[1].position_end == 19
    assert parse_tree[2].position_end == 27
    assert parse_tree.position_end == 27


def test_pos_to_linecol():

    def grammar(): return ("a", "b", "c")

    parser = ParserPython(grammar)

    parse_tree = parser.parse("a\n\n\n b\nc")

    a_pos = parse_tree[0].position
    assert parser.pos_to_linecol(a_pos) == (1, 1)
    b_pos = parse_tree[1].position
    assert parser.pos_to_linecol(b_pos) == (4, 2)
    c_pos = parse_tree[2].position
    assert parser.pos_to_linecol(c_pos) == (5, 1)



"""
Various utilities.
"""

# isstr check if object is of string type.
# This works for both python 2 and 3
# Taken from http://stackoverflow.com/questions/11301138/how-to-check-if-variable-is-string-with-python-2-and-3-compatibility
try:
    basestring  # attempt to evaluate basestring
    def isstr(s):
        return isinstance(s, basestring)
except NameError:
    def isstr(s):
        return isinstance(s, str)


# -*- coding: utf-8 -*-
#######################################################################
# Name: test_examples
# Purpose: Test that examples run without errors.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014-2015 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
import pytest   # noqa
import os
import sys
import glob

PY_LT_3_5 = sys.version_info < (3, 5)
if PY_LT_3_5:
    import imp
else:
    import importlib


def test_examples():

    examples_folder = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                   '..', '..', 'examples')
    if not os.path.exists(examples_folder):
        print('Warning: Examples not found. Skipping tests.')
        return

    examples_pat = os.path.join(examples_folder, '*', '*.py')

    # Filter out __init__.py
    examples = [f for f in glob.glob(examples_pat) if f != '__init__.py']
    for e in examples:
        example_dir = os.path.dirname(e)
        sys.path.insert(0, example_dir)
        (module_name, _) = os.path.splitext(os.path.basename(e))
        if PY_LT_3_5:
            (module_file, module_path, desc) = \
                imp.find_module(module_name, [example_dir])
            mod = imp.load_module(module_name, module_file, module_path, desc)
        else:
            mod_spec = importlib.util.spec_from_file_location(module_name, e)
            mod = importlib.util.module_from_spec(mod_spec)
            mod_spec.loader.exec_module(mod)

        if hasattr(mod, 'main'):
            mod.main(debug=False)



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_flags
# Purpose: Test for parser flags
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import re
import pytest

# Grammar
from arpeggio import ParserPython, Optional, EOF
from arpeggio import RegExMatch as _
from arpeggio import NoMatch


def foo():      return 'r', bar, Optional(qux), baz, Optional(ham), Optional(buz), EOF
def bar():      return 'BAR'
def baz():      return _(r'1\w+')
def buz():      return _(r'Aba*', ignore_case=True)
def qux():      return _(r'/\*.*\*/', multiline=True)
def ham():      return _(r'/\*.*\*/', re_flags=re.DOTALL)  # equivalent to qux


@pytest.fixture
def parser_ci():
    return ParserPython(foo, ignore_case=True)


@pytest.fixture
def parser_nonci():
    return ParserPython(foo, ignore_case=False)


def test_parse_tree_ci(parser_ci):
    input_str = "R bar 1baz"
    parse_tree = parser_ci.parse(input_str)
    assert parse_tree is not None


def test_parse_tree_nonci(parser_nonci):
    input_str = "R bar 1baz"
    with pytest.raises(NoMatch):
        parser_nonci.parse(input_str)


def test_flags_override(parser_nonci):
    # Parser is not case insensitive
    # But the buz match is.
    input_str = "r BAR 1baz abaaaaAAaaa"
    parse_tree = parser_nonci.parse(input_str)
    assert parse_tree is not None


def test_multiline_comment(parser_nonci):
    input_str = "r BAR /*1baz\nabaaaaAAaaa\n*/1baz"
    parse_tree = parser_nonci.parse(input_str)
    assert parse_tree is not None


def test_multiline_comment_by_re_flags(parser_nonci):
    input_str = "r BAR 1baz/*this\nis\nnot\nparsed*/"
    parse_tree = parser_nonci.parse(input_str)
    assert parse_tree is not None



import os
from arpeggio.cleanpeg import ParserPEG


def test_issue_22():
    """
    Infinite recursion during resolving of a grammar given in a clean PEG
    notation.
    """
    current_dir = os.path.dirname(__file__)

    grammar1 = open(os.path.join(current_dir, 'grammar1.peg')).read()
    parser1 = ParserPEG(grammar1, 'belang')
    parser1.parse('a [0]')
    parser1.parse('a (0)')

    grammar2 = open(os.path.join(current_dir, 'grammar2.peg')).read()
    parser2 = ParserPEG(grammar2, 'belang', debug=True)
    parser2.parse('a [0](1)[2]')



"""
See https://github.com/textX/Arpeggio/issues/43
"""
from arpeggio.cleanpeg import ParserPEG


def try_grammer(peg):
    p = ParserPEG(peg, 'letters', debug=False)
    p.parse(""" { a b } """)
    p.parse(""" { b a } """)


def test_plain_grammar():
    try_grammer("""
    letters = "{" ("a" "b")# "}"
    n = "9"
    """)


def test_bs_at_eol():
    try_grammer("""
    letters = "{" ("a" "b")# "}" \
    n = "9"
    """)


def test_move_unordered_group_to_last_line_in_grammar():
    try_grammer("""
    n = "9"
    letters = "{" ("a" "b")# "}" \
    """)



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_unicode
# Purpose: Tests matching unicode characters
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest  # noqa

# Grammar
from arpeggio import ParserPython

def grammar():      return first, "±", second
def first():        return "♪"
def second():       return "a"


def test_unicode_match():
    parser = ParserPython(grammar)

    parse_tree = parser.parse("♪ ± a")
    assert parse_tree



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_parser_params
# Purpose: Test for parser parameters.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest
from arpeggio import ParserPython, NoMatch
import sys


def test_autokwd():
    """
    autokwd will match keywords on word boundaries.
    """
    def grammar():
        return ("one", "two", "three")

    parser = ParserPython(grammar, autokwd=True)

    # If autokwd is enabled this should parse without error.
    parser.parse("one two three")

    # But this will not parse because each word to match
    # will be, by default, tried to match as a whole word
    with pytest.raises(NoMatch):
        parser.parse("onetwothree")

    parser = ParserPython(grammar, autokwd=False)
    # If we turn off the autokwd than this will match.
    parser.parse("one two three")
    parser.parse("onetwothree")


def test_skipws():
    """
    skipws will skip whitespaces.
    """

    def grammar():
        return ("one", "two", "three")

    parser = ParserPython(grammar)

    # If skipws is on this should parse without error.
    parser.parse("one two three")

    # If not the same input will raise exception.
    parser = ParserPython(grammar, skipws=False)
    with pytest.raises(NoMatch):
        parser.parse("one two three")


def test_ws():
    """
    ws consists of chars that will be skipped if skipws is enables.
    By default it consists of space, tab and newline.
    """

    def grammar():
        return ("one", "two", "three")

    parser = ParserPython(grammar)

    # With default ws this should parse without error
    parser.parse("""one
                  two   three""")

    # If we make only a space char to be ws than the
    # same input will raise exception.
    parser = ParserPython(grammar, ws=" ")
    with pytest.raises(NoMatch):
        parser.parse("""one
                  two   three""")

    # But if only spaces are between words than it will
    # parse.
    parser.parse("one two  three")


def test_file(capsys):
    """
    'file' specifies an output file for the DebugPrinter mixin.
    """

    def grammar():
        return ("one", "two", "three")

    # First use stdout
    parser = ParserPython(grammar, debug=True, file=sys.stdout)
    out, err = capsys.readouterr()

    parser.dprint('this is stdout')
    out, err = capsys.readouterr()
    assert out == 'this is stdout\n'
    assert err == ''

    # Now use stderr
    parser = ParserPython(grammar, debug=False, file=sys.stderr)
    out, err = capsys.readouterr()

    parser.dprint('this is stderr')
    out, err = capsys.readouterr()
    assert out == ''
    assert err == 'this is stderr\n'



from __future__ import unicode_literals
import pytest
from arpeggio import ParserPython, UnorderedGroup, Optional, \
    EOF, NoMatch


def test_nondeterministic_unordered_group():

    def root():
        return 'word1', UnorderedGroup(some_rule, 'word2', some_rule), EOF

    def some_rule():
        return Optional('word2'), Optional('word3')

    content = '''word1 word2 '''

    # If the 'word2' from unordered group in the `root` rule matches first
    # the input parses, else it fails.
    # We repeat parser construction and parsing many times to check
    # if it fails every time. The current fix will iterate in order from left
    # to right and repeat matching until all rules in a unordered group
    # succeeds.
    fail = 0
    success = 0
    for _ in range(100):
        try:
            parser = ParserPython(root)
            parser.parse(content)
            success += 1
        except NoMatch:
            fail += 1

    assert fail == 100



#######################################################################
# Name: robot.py
# Purpose: Simple DSL for defining robot movement.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2011-2014 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# This example is inspired by an example from LISA tool (http://labraj.uni-mb.si/lisa/)
# presented during the lecture given by prof. Marjan Mernik (http://lpm.uni-mb.si/mernik/)
# at the University of Novi Sad in June, 2011.
#
# An example of the robot program:
#    begin
#        up
#        up
#        left
#        down
#        right
#    end
#######################################################################
from __future__ import print_function, unicode_literals

import os
from arpeggio import ZeroOrMore, EOF, PTNodeVisitor, ParserPython, \
    visit_parse_tree
from arpeggio.export import PMDOTExporter, PTDOTExporter

# Grammar rules
def robot():      return 'begin', ZeroOrMore(command), 'end', EOF
def command():      return [UP, DOWN, LEFT, RIGHT]
def UP():           return 'up'
def DOWN():         return 'down'
def LEFT():         return 'left'
def RIGHT():        return 'right'


# Semantic actions visitor
class RobotVisitor(PTNodeVisitor):

    def visit_robot(self, node, children):
        if self.debug:
            print("Evaluating position")
        position = [0, 0]
        for move in children:
            position[0] += move[0]
            position[1] += move[1]
        return position

    def visit_command(self, node, children):
        if self.debug:
            print("Command")
        return children[0]

    def visit_UP(self, node, children):
        if self.debug:
            print("Going up")
        return (0, 1)

    def visit_DOWN(self, node, children):
        if self.debug:
            print("Going down")
        return (0, -1)

    def visit_LEFT(self, node, children):
        if self.debug:
            print("Going left")
        return (-1, 0)

    def visit_RIGHT(self, node, children):
        if self.debug:
            print("Going right")
        return (1, 0)


def main(debug=False):

    # Load program
    current_dir = os.path.dirname(__file__)
    input_program = open(os.path.join(current_dir, 'program.rbt'), 'r').read()

    # First we will make a parser - an instance of the robot parser model.
    # Parser model is given in the form of python constructs therefore we
    # are using ParserPython class.
    parser = ParserPython(robot, debug=debug)

    # We create a parse tree out of textual input
    parse_tree = parser.parse(input_program)

    # visit_parse_tree will start semantic analysis.
    # In this case semantic analysis will evaluate expression and
    # returned value will be the final position of the robot.
    result = visit_parse_tree(parse_tree, RobotVisitor(debug=debug))

    if debug:
        print("position = ", result)


if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    main(debug=True)



##############################################################################
# Name: csv.py
# Purpose: Implementation of CSV parser in arpeggio.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
##############################################################################

from __future__ import unicode_literals
import pprint
import os
from arpeggio import *
from arpeggio import RegExMatch as _

def record():                   return field, ZeroOrMore(",", field)
def field():                    return [quoted_field, field_content]
def quoted_field():             return '"', field_content_quoted, '"'
def field_content():            return _(r'([^,\n])+')
def field_content_quoted():     return _(r'(("")|([^"]))+')
def csvfile():                  return OneOrMore([record, '\n']), EOF


class CSVVisitor(PTNodeVisitor):
    def visit_field(self, node, children):
        value = children[0]
        try:
            return float(value)
        except:
            pass
        try:
            return int(value)
        except:
            return value

    def visit_record(self, node, children):
        # record is a list of fields. The children nodes are fields so just
        # transform it to list.
        return list(children)

    def visit_csvfile(self, node, children):
        # We are not interested in newlines so we will filter them.
        return [x for x in children if x!='\n']



def main(debug=False):
    # First we will make a parser - an instance of the CVS parser model.
    # Parser model is given in the form of python constructs therefore we
    # are using ParserPython class.
    # Skipping of whitespace will be done only for tabs and spaces. Newlines
    # have semantics in csv files. They are used to separate records.
    parser = ParserPython(csvfile, ws='\t ', debug=debug)

    # Creating parse tree out of textual input
    current_dir = os.path.dirname(__file__)
    test_data = open(os.path.join(current_dir, 'test_data.csv'), 'r').read()
    parse_tree = parser.parse(test_data)

    # Create list of lists using visitor
    csv_content = visit_parse_tree(parse_tree, CSVVisitor())
    print("This is a list of lists with the data from CSV file.")
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(csv_content)

if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    main(debug=True)




#######################################################################
# Name: calc_peg.py
# Purpose: Simple expression evaluator example using PEG language
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2009-2015 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# This example is functionally equivalent to calc.py. The difference is that
# in this example grammar is specified using PEG language instead of python constructs.
# Semantic actions are used to calculate expression during semantic
# analysis.
# Parser model as well as parse tree exported to dot files should be
# the same as parser model and parse tree generated in calc.py example.
#######################################################################
from __future__ import absolute_import, unicode_literals, print_function

import os
from arpeggio.cleanpeg import ParserPEG
from arpeggio import visit_parse_tree
from calc import CalcVisitor


def main(debug=False):

    # Grammar is defined using textual specification based on PEG language.
    # Load grammar form file.
    calc_grammar = open(os.path.join(os.path.dirname(__file__),
                                     'calc_clean.peg'), 'r').read()

    # First we will make a parser - an instance of the calc parser model.
    # Parser model is given in the form of PEG notation therefore we
    # are using ParserPEG class. Root rule name (parsing expression) is "calc".
    parser = ParserPEG(calc_grammar, "calc", debug=debug)

    # An expression we want to evaluate
    input_expr = "-(4-1)*5+(2+4.67)+5.89/(.2+7)"

    # Then parse tree is created out of the input_expr expression.
    parse_tree = parser.parse(input_expr)

    # The result is obtained by semantic evaluation using visitor class.
    # visit_parse_tree will start semantic analysis.
    # In this case semantic analysis will evaluate expression and
    # returned value will be evaluated result of the input_expr expression.
    result = visit_parse_tree(parse_tree, CalcVisitor(debug=debug))

    # Check that result is valid
    assert (result - -7.51194444444) < 0.0001

    print("{} = {}".format(input_expr, result))

if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    main(debug=True)




##############################################################################
# Name: csv.py
# Purpose: Implementation of CSV parser in arpeggio.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014-2015 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
##############################################################################

import os
import pprint
from arpeggio import visit_parse_tree
from arpeggio.cleanpeg import ParserPEG
from examples.csv.csv import CSVVisitor

def main(debug=False):
    # First we will make a parser - an instance of the CVS parser model.
    # Parser model is given in the form of clean PEG description therefore we
    # are using ParserPEG class from arpeggio.clenapeg.  Grammar is loaded from
    # csv.peg file Skipping of whitespace will be done only for tabs and
    # spaces. Newlines have semantics in csv files. They are used to separate
    # records.
    current_dir = os.path.dirname(__file__)
    csv_grammar = open(os.path.join(current_dir, 'csv.peg'), 'r').read()
    parser = ParserPEG(csv_grammar, 'csvfile', ws='\t ', debug=debug)

    # Creating parse tree out of textual input
    test_data = open(os.path.join(current_dir, 'test_data.csv'), 'r').read()
    parse_tree = parser.parse(test_data)

    # Create list of lists using visitor
    csv_content = visit_parse_tree(parse_tree, CSVVisitor())
    print("This is a list of lists with the data from CSV file.")
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(csv_content)

if __name__ == "__main__":
    # In debug mode dot (graphviz) files for parser model
    # and parse tree will be created for visualization.
    # Checkout current folder for .dot files.
    main(debug=True)




# -*- coding: utf-8 -*-
#######################################################################
# Name: test_peg_parser
# Purpose: Test for parser constructed using PEG textual grammars.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014-2017 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
import pytest  # noqa
from arpeggio import Sequence, NonTerminal, NoMatch
from arpeggio.peg import ParserPEG
from arpeggio.cleanpeg import ParserPEG as ParserPEGClean

grammar = r'''
    // This is a comment
    number <- r'\d*\.\d*|\d+';
    factor <- ("+" / "-")?
              (number / "(" expression ")");
    // This is another comment
    term <- factor (( "*" / "/") factor)*;
    expression <- term (("+" / "-") term)*;
    calc <- expression+ EOF;
    // And final comment at the end of file
'''

clean_grammar = grammar.replace('<-', '=').replace(';', '')

@pytest.mark.parametrize('parser', [ParserPEG(grammar, 'calc'),
                                    ParserPEGClean(clean_grammar, 'calc')])
def test_construct_parser(parser):
    assert parser.parser_model.rule_name == 'calc'
    assert isinstance(parser.parser_model, Sequence)
    assert parser.parser_model.nodes[0].name == 'OneOrMore'


@pytest.mark.parametrize('parser', [ParserPEG(grammar, 'calc'),
                                    ParserPEGClean(clean_grammar, 'calc')])
def test_parse_input(parser):

    input = "4+5*7/3.45*-45*(2.56+32)/-56*(2-1.34)"
    result = parser.parse(input)

    assert isinstance(result, NonTerminal)
    assert str(result) == "4 | + | 5 | * | 7 | / | 3.45 | * | - | 45 | * | ( | 2.56 | + | 32 | ) | / | - | 56 | * | ( | 2 | - | 1.34 | ) | "  # noqa
    assert repr(result) == "[ [ [ [ number '4' [0] ] ],  '+' [1], [ [ number '5' [2] ],  '*' [3], [ number '7' [4] ],  '/' [5], [ number '3.45' [6] ],  '*' [10], [  '-' [11], number '45' [12] ],  '*' [14], [  '(' [15], [ [ [ number '2.56' [16] ] ],  '+' [20], [ [ number '32' [21] ] ] ],  ')' [23] ],  '/' [24], [  '-' [25], number '56' [26] ],  '*' [28], [  '(' [29], [ [ [ number '2' [30] ] ],  '-' [31], [ [ number '1.34' [32] ] ] ],  ')' [36] ] ] ], EOF [37] ]"  # noqa


@pytest.mark.parametrize('parser',
                         [ParserPEG(grammar, 'calc', reduce_tree=True),
                          ParserPEGClean(clean_grammar, 'calc',
                                         reduce_tree=True)])
def test_reduce_tree(parser):

    input = "4+5*7/3.45*-45*(2.56+32)/-56*(2-1.34)"
    result = parser.parse(input)

    assert isinstance(result, NonTerminal)

    assert str(result) == "4 | + | 5 | * | 7 | / | 3.45 | * | - | 45 | * | ( | 2.56 | + | 32 | ) | / | - | 56 | * | ( | 2 | - | 1.34 | ) | "  # noqa
    assert repr(result) == "[ [ number '4' [0],  '+' [1], [ number '5' [2],  '*' [3], number '7' [4],  '/' [5], number '3.45' [6],  '*' [10], [  '-' [11], number '45' [12] ],  '*' [14], [  '(' [15], [ number '2.56' [16],  '+' [20], number '32' [21] ],  ')' [23] ],  '/' [24], [  '-' [25], number '56' [26] ],  '*' [28], [  '(' [29], [ number '2' [30],  '-' [31], number '1.34' [32] ],  ')' [36] ] ] ], EOF [37] ]"  # noqa


def test_unordered_group():

    grammar = """
    g <- (("a" "b") "c" )#;
    """

    parser = ParserPEG(grammar, 'g', reduce_tree=True)

    r = parser.parse("c a b")
    assert isinstance(r, NonTerminal)

    r = parser.parse("a b c")
    assert isinstance(r, NonTerminal)

    with pytest.raises(NoMatch):
        parser.parse("a c b")



# -*- coding: utf-8 -*-
#######################################################################
# Name: cleanpeg.py
# Purpose: This module is a variation of the original peg.py.
#   The syntax is slightly changed to be more readable and familiar to
#   python users. It is based on the Yash's suggestion - issue 11
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014-2017 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import print_function, unicode_literals

from arpeggio import Optional, ZeroOrMore, Not, OneOrMore, EOF, ParserPython, \
    visit_parse_tree
from arpeggio import RegExMatch as _
from .peg import PEGVisitor
from .peg import ParserPEG as ParserPEGOrig

__all__ = ['ParserPEG']

# Lexical invariants
ASSIGNMENT = "="
ORDERED_CHOICE = "/"
ZERO_OR_MORE = "*"
ONE_OR_MORE = "+"
OPTIONAL = "?"
UNORDERED_GROUP = "#"
AND = "&"
NOT = "!"
OPEN = "("
CLOSE = ")"

# PEG syntax rules
def peggrammar():       return OneOrMore(rule), EOF
def rule():             return rule_name, ASSIGNMENT, ordered_choice
def ordered_choice():   return sequence, ZeroOrMore(ORDERED_CHOICE, sequence)
def sequence():         return OneOrMore(prefix)
def prefix():           return Optional([AND, NOT]), sufix
def sufix():            return expression, Optional([OPTIONAL,
                                                     ZERO_OR_MORE,
                                                     ONE_OR_MORE,
                                                     UNORDERED_GROUP])
def expression():       return [regex, rule_crossref,
                                (OPEN, ordered_choice, CLOSE),
                                str_match], Not(ASSIGNMENT)

# PEG Lexical rules
def regex():            return [("r'", _(r'''[^'\\]*(?:\\.[^'\\]*)*'''), "'"),
                                ('r"', _(r'''[^"\\]*(?:\\.[^"\\]*)*'''), '"')]
def rule_name():        return _(r"[a-zA-Z_]([a-zA-Z_]|[0-9])*")
def rule_crossref():    return rule_name
def str_match():        return _(r'''(?s)('[^'\\]*(?:\\.[^'\\]*)*')|'''
                                 r'''("[^"\\]*(?:\\.[^"\\]*)*")''')
def comment():          return "//", _(".*\n")


class ParserPEG(ParserPEGOrig):

    def _from_peg(self, language_def):
        parser = ParserPython(peggrammar, comment, reduce_tree=False,
                              debug=self.debug)
        parser.root_rule_name = self.root_rule_name
        parse_tree = parser.parse(language_def)

        return visit_parse_tree(parse_tree, PEGVisitor(self.root_rule_name,
                                                       self.comment_rule_name,
                                                       self.ignore_case,
                                                       debug=self.debug))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
# Name: arpeggio.py
# Purpose: PEG parser interpreter
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#
# Arpeggio is an implementation of packrat parser interpreter based on PEG
# grammars.
# Parsers are defined using python language construction or PEG language.
###############################################################################

from io import open
import os
import sys
from setuptools import setup

VERSIONFILE = "arpeggio/__init__.py"
VERSION = None
for line in open(VERSIONFILE, "r", encoding='utf8').readlines():
    if line.startswith('__version__'):
        VERSION = line.split('"')[1]

if not VERSION:
    raise RuntimeError('No version defined in arpeggio/__init__.py')

if sys.argv[-1].startswith('publish'):
    if os.system("pip list | grep wheel"):
        print("wheel not installed.\nUse `pip install wheel`.\nExiting.")
        sys.exit()
    if os.system("pip list | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    if sys.argv[-1] == 'publishtest':
        os.system("twine upload -r test dist/*")
    else:
        os.system("twine upload dist/*")
        print("You probably want to also tag the version now:")
        print("  git tag -a {0} -m 'version {0}'".format(VERSION))
        print("  git push --tags")
    sys.exit()

setup(version=VERSION)



# -*- coding: utf-8 -*-
#######################################################################
# Name: test_parsing_expressions
# Purpose: Test for parsing expressions.
# Author: Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2014-2017 Igor R. Dejanović <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################

from __future__ import unicode_literals
import pytest
from arpeggio import ParserPython, UnorderedGroup, ZeroOrMore, OneOrMore, \
    NoMatch, EOF, Optional, And, Not, StrMatch, RegExMatch


def test_sequence():

    def grammar():
        return ("a", "b", "c")

    parser = ParserPython(grammar)

    parsed = parser.parse("a b c")

    assert str(parsed) == "a | b | c"
    assert repr(parsed) == "[  'a' [0],  'b' [2],  'c' [4] ]"


def test_ordered_choice():

    def grammar():
        return ["a", "b", "c"], EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("b")

    assert str(parsed) == "b | "
    assert repr(parsed) == "[  'b' [0], EOF [1] ]"

    parsed = parser.parse("c")
    assert str(parsed) == "c | "
    assert repr(parsed) == "[  'c' [0], EOF [1] ]"

    with pytest.raises(NoMatch):
        parser.parse("ab")

    with pytest.raises(NoMatch):
        parser.parse("bb")


def test_unordered_group():

    def grammar():
        return UnorderedGroup("a", "b", "c"), EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("b a c")

    assert str(parsed) == "b | a | c | "
    assert repr(parsed) == "[  'b' [0],  'a' [2],  'c' [4], EOF [5] ]"

    with pytest.raises(NoMatch):
        parser.parse("a b a c")

    with pytest.raises(NoMatch):
        parser.parse("a c")

    with pytest.raises(NoMatch):
        parser.parse("b b a c")


def test_unordered_group_with_separator():

    def grammar():
        return UnorderedGroup("a", "b", "c", sep=StrMatch(",")), EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("b, a , c")

    assert str(parsed) == "b | , | a | , | c | "
    assert repr(parsed) == \
        "[  'b' [0],  ',' [1],  'a' [3],  ',' [5],  'c' [7], EOF [8] ]"

    with pytest.raises(NoMatch):
        parser.parse("a, b, a, c")

    with pytest.raises(NoMatch):
        parser.parse("a, c")

    with pytest.raises(NoMatch):
        parser.parse("b, b, a, c")

    with pytest.raises(NoMatch):
        parser.parse(",a, b, c")

    with pytest.raises(NoMatch):
        parser.parse("a, b, c,")

    with pytest.raises(NoMatch):
        parser.parse("a, ,b, c")


def test_unordered_group_with_optionals():

    def grammar():
        return UnorderedGroup("a", Optional("b"), "c"), EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("b a c")
    assert str(parsed) == "b | a | c | "

    parsed = parser.parse("a c b")
    assert str(parsed) == "a | c | b | "

    parsed = parser.parse("a c")
    assert str(parsed) == "a | c | "

    with pytest.raises(NoMatch):
        parser.parse("a b c b")

    with pytest.raises(NoMatch):
        parser.parse("a b ")


def test_unordered_group_with_optionals_and_separator():

    def grammar():
        return UnorderedGroup("a", Optional("b"), "c", sep=","), EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("b, a, c")
    assert parsed

    parsed = parser.parse("a, c, b")
    assert parsed

    parsed = parser.parse("a, c")
    assert parsed

    with pytest.raises(NoMatch):
        parser.parse("a, b, c, b")

    with pytest.raises(NoMatch):
        parser.parse("a, b ")

    with pytest.raises(NoMatch):
        parser.parse("a, c, ")

    with pytest.raises(NoMatch):
        parser.parse("a, b c ")

    with pytest.raises(NoMatch):
        parser.parse(",a, c ")


def test_zero_or_more():

    def grammar():
        return ZeroOrMore("a"), EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("aaaaaaa")

    assert str(parsed) == "a | a | a | a | a | a | a | "
    assert repr(parsed) == "[  'a' [0],  'a' [1],  'a' [2],"\
        "  'a' [3],  'a' [4],  'a' [5],  'a' [6], EOF [7] ]"

    parsed = parser.parse("")

    assert str(parsed) == ""
    assert repr(parsed) == "[ EOF [0] ]"

    with pytest.raises(NoMatch):
        parser.parse("bbb")


def test_zero_or_more_with_separator():

    def grammar():
        return ZeroOrMore("a", sep=","), EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("a, a , a , a ,  a,a, a")

    assert str(parsed) == \
        "a | , | a | , | a | , | a | , | a | , | a | , | a | "
    assert repr(parsed) == \
        "[  'a' [0],  ',' [1],  'a' [3],  ',' [5],  'a' [7],  ',' [9],  "\
        "'a' [11],  ',' [13],  'a' [16],  ',' [17],  'a' [18],  ',' [19],"\
        "  'a' [21], EOF [22] ]"

    parsed = parser.parse("")

    assert str(parsed) == ""
    assert repr(parsed) == "[ EOF [0] ]"

    with pytest.raises(NoMatch):
        parser.parse("aa a")

    with pytest.raises(NoMatch):
        parser.parse(",a,a ,a")

    with pytest.raises(NoMatch):
        parser.parse("a,a ,a,")

    with pytest.raises(NoMatch):
        parser.parse("bbb")


def test_zero_or_more_with_optional_separator():

    def grammar():
        return ZeroOrMore("a", sep=RegExMatch(",?")), EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("a, a , a   a ,  a,a, a")

    assert str(parsed) == \
        "a | , | a | , | a | a | , | a | , | a | , | a | "
    assert repr(parsed) == \
        "[  'a' [0],  ',' [1],  'a' [3],  ',' [5],  'a' [7],  "\
        "'a' [11],  ',' [13],  'a' [16],  ',' [17],  'a' [18],  ',' [19],"\
        "  'a' [21], EOF [22] ]"

    parsed = parser.parse("")

    assert str(parsed) == ""
    assert repr(parsed) == "[ EOF [0] ]"

    parser.parse("aa a")

    with pytest.raises(NoMatch):
        parser.parse(",a,a ,a")

    with pytest.raises(NoMatch):
        parser.parse("a,a ,a,")

    with pytest.raises(NoMatch):
        parser.parse("bbb")


def test_one_or_more():

    def grammar():
        return OneOrMore("a"), "b"

    parser = ParserPython(grammar)

    parsed = parser.parse("aaaaaa a  b")

    assert str(parsed) == "a | a | a | a | a | a | a | b"
    assert repr(parsed) == "[  'a' [0],  'a' [1],  'a' [2],"\
        "  'a' [3],  'a' [4],  'a' [5],  'a' [7],  'b' [10] ]"

    parser.parse("ab")

    with pytest.raises(NoMatch):
        parser.parse("")

    with pytest.raises(NoMatch):
        parser.parse("b")


def test_one_or_more_with_separator():

    def grammar():
        return OneOrMore("a", sep=","), "b"

    parser = ParserPython(grammar)

    parsed = parser.parse("a, a, a, a  b")

    assert str(parsed) == "a | , | a | , | a | , | a | b"
    assert repr(parsed) == \
        "[  'a' [0],  ',' [1],  'a' [3],  ',' [4],  'a' [6],  ',' [7],  "\
        "'a' [9],  'b' [12] ]"

    parser.parse("a b")

    with pytest.raises(NoMatch):
        parser.parse("")

    with pytest.raises(NoMatch):
        parser.parse("b")

    with pytest.raises(NoMatch):
        parser.parse("a a b")

    with pytest.raises(NoMatch):
        parser.parse("a a, b")

    with pytest.raises(NoMatch):
        parser.parse(", a, a b")


def test_one_or_more_with_optional_separator():

    def grammar():
        return OneOrMore("a", sep=RegExMatch(",?")), "b"

    parser = ParserPython(grammar)

    parsed = parser.parse("a, a  a, a  b")

    assert str(parsed) == "a | , | a | a | , | a | b"
    assert repr(parsed) == \
        "[  'a' [0],  ',' [1],  'a' [3],  'a' [6],  ',' [7],  "\
        "'a' [9],  'b' [12] ]"

    parser.parse("a b")

    with pytest.raises(NoMatch):
        parser.parse("")

    with pytest.raises(NoMatch):
        parser.parse("b")

    with pytest.raises(NoMatch):
        parser.parse("a a, b")

    with pytest.raises(NoMatch):
        parser.parse(", a, a b")


def test_optional():

    def grammar():
        return Optional("a"), "b", EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("ab")

    assert str(parsed) == "a | b | "
    assert repr(parsed) == "[  'a' [0],  'b' [1], EOF [2] ]"

    parsed = parser.parse("b")

    assert str(parsed) == "b | "
    assert repr(parsed) == "[  'b' [0], EOF [1] ]"

    with pytest.raises(NoMatch):
        parser.parse("aab")

    with pytest.raises(NoMatch):
        parser.parse("")


# Syntax predicates

def test_and():

    def grammar():
        return "a", And("b"), ["c", "b"], EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("ab")
    assert str(parsed) == "a | b | "
    assert repr(parsed) == "[  'a' [0],  'b' [1], EOF [2] ]"

    # 'And' will try to match 'b' and fail so 'c' will never get matched
    with pytest.raises(NoMatch):
        parser.parse("ac")

    # 'And' will not consume 'b' from the input so second 'b' will never match
    with pytest.raises(NoMatch):
        parser.parse("abb")


def test_not():

    def grammar():
        return "a", Not("b"), ["b", "c"], EOF

    parser = ParserPython(grammar)

    parsed = parser.parse("ac")

    assert str(parsed) == "a | c | "
    assert repr(parsed) == "[  'a' [0],  'c' [1], EOF [2] ]"

    # Not will will fail on 'b'
    with pytest.raises(NoMatch):
        parser.parse("ab")

    # And will not consume 'c' from the input so 'b' will never match
    with pytest.raises(NoMatch):
        parser.parse("acb")
